import functools
import threading
from typing import Callable, Literal

from airalogy.assigner.assigner_result import AssignerResult


def flatten_list(lst):
    result = []
    for i in lst:
        if isinstance(i, list):
            result.extend(flatten_list(i))
        else:
            result.append(i)
    return result


def unique_list(lst):
    return list(set(lst))


AssignerMode = Literal["manual", "auto_first", "auto", "auto_force"]


class AssignerBase:
    _lock = threading.Lock()
    assigned_info: dict[str, tuple[list[str], Callable, AssignerMode]] = {}
    dependent_info: dict[str, list[tuple[str, Callable, AssignerMode]]] = {}

    def __init_subclass__(cls, **kwargs):
        with AssignerBase._lock:
            cls.assigned_info = AssignerBase.assigned_info
            cls.dependent_info = AssignerBase.dependent_info
            AssignerBase.assigned_info = {}
            AssignerBase.dependent_info = {}

    @classmethod
    def assigner(
        cls,
        assigned_fields: list[str],
        dependent_fields: list[str],
        mode: AssignerMode = "auto_first",
    ):
        def decorator(assign_func: Callable):
            if len(assigned_fields) == 0:
                raise ValueError(
                    f"assigned_fields must be not empty when using {assign_func.__name__}."
                )
            if len(dependent_fields) == 0 and mode != "manual":
                raise ValueError(
                    f"dependent_fields must be not empty when using {assign_func.__name__} in mode {mode}."
                )
            for key in assigned_fields:
                if key in cls.assigned_info:
                    raise ValueError(
                        f"assigned_fields: {key} has been defined in other assigner."
                    )
                cls.assigned_info[key] = (dependent_fields, assign_func, mode)
            for key in dependent_fields:
                if key not in cls.dependent_info:
                    cls.dependent_info[key] = []
                for assigned_key in assigned_fields:
                    cls.dependent_info[key].append((assigned_key, assign_func, mode))

            @functools.wraps(assign_func)
            def wrapper(dependent_data: dict) -> AssignerResult:
                # check dependent_data type
                if not isinstance(dependent_data, dict):
                    return AssignerResult(
                        success=False,
                        assigned_fields=None,
                        error_message=f"The parameter of {assign_func.__name__} must be a dict type.",
                    )
                # 检查 dependent data 是否包含所有 dependent_fields
                missing_keys = [
                    key for key in dependent_fields if key not in dependent_data
                ]
                if len(missing_keys) > 0:
                    return AssignerResult(
                        success=False,
                        assigned_fields=None,
                        error_message=f"Missing dependent rfs: {missing_keys} for assigned_fields: {assigned_fields}, in {assign_func.__name__}",
                    )

                try:
                    result = assign_func(dependent_data)
                except Exception as e:
                    return AssignerResult(
                        success=False,
                        assigned_fields=None,
                        error_message=str(e),
                    )

                # 检查 assign_func 的返回值
                if not isinstance(result, AssignerResult):
                    return AssignerResult(
                        success=False,
                        assigned_fields=None,
                        error_message=f"The return value of {assign_func.__name__} must be a AssignerResult.",
                    )
                if result.success:
                    # 检查返回的 dict 是否包含所有 assigned_fields
                    missing_keys = [
                        key
                        for key in assigned_fields
                        if key not in result.assigned_fields
                    ]
                    if len(missing_keys) > 0:
                        return AssignerResult(
                            success=False,
                            assigned_fields=None,
                            error_message=f"Missing assigned rfs: {missing_keys} in the return value of {assign_func.__name__}",
                        )

                return result

            return staticmethod(wrapper)

        return decorator

    @classmethod
    def get_dependent_fields_of_assigned_key(cls, assigned_key: str) -> list[str]:
        if assigned_key in cls.assigned_info:
            dependent_keys = cls.assigned_info[assigned_key][0]
            for key in dependent_keys:
                dependent_keys.extend(cls.get_dependent_fields_of_assigned_key(key))

            return unique_list(flatten_list(dependent_keys))
        else:
            return []

    @classmethod
    def get_assigned_fields_of_dependent_key(cls, dependent_key: str) -> list[str]:
        if dependent_key in cls.dependent_info:
            assigned_fields = []
            for item in cls.dependent_info[dependent_key]:
                key = item[0]
                assigned_fields.append(key)
                assigned_fields.extend(cls.get_assigned_fields_of_dependent_key(key))

            return unique_list(flatten_list(assigned_fields))
        else:
            return []

    @classmethod
    def get_assign_func_of_assigned_key(cls, assigned_key: str) -> Callable | None:
        if assigned_key in cls.assigned_info:
            return cls.assigned_info[assigned_key][1]
        else:
            return None

    @classmethod
    def get_assign_funcs_of_dependent_key(cls, dependent_key: str) -> list[Callable]:
        if dependent_key in cls.dependent_info:
            return [item[1] for item in cls.dependent_info[dependent_key]]
        else:
            return []

    @classmethod
    def all_assigned_fields(cls) -> dict[str, list[str]]:
        return {
            k: {
                "dependent_fields": cls.get_dependent_fields_of_assigned_key(k),
                "mode": v[2],
            }
            for k, v in cls.assigned_info.items()
        }

    @classmethod
    def assign(cls, rf_name: str, dependent_data: dict) -> AssignerResult:
        dep_rfs = cls.get_dependent_fields_of_assigned_key(rf_name)
        for rf in dep_rfs:
            info = cls.assigned_info.get(rf)
            if info and info[2] != "auto_first":
                res = cls.assign(rf, dependent_data)
                if res.success:
                    dependent_data[rf] = res.assigned_fields[rf]
                else:
                    return res
            if rf not in dependent_data:
                return AssignerResult(
                    success=False,
                    assigned_fields=None,
                    error_message=f"Missing dependent rf: {rf} for assigned rf: {rf_name}",
                )

        assign_func = cls.get_assign_func_of_assigned_key(rf_name)
        if assign_func is None:
            return AssignerResult(
                success=False,
                assigned_fields=None,
                error_message=f"Cannot find assign function for rf: {rf_name}",
            )

        return assign_func(dependent_data)


def assigner(
    assigned_fields: list[str],
    dependent_fields: list[str],
    mode: AssignerMode = "auto_first",
):
    return AssignerBase.assigner(assigned_fields, dependent_fields, mode)



"""
Built-in type: AiralogyMarkdown
"""

from pydantic import Field
from typing_extensions import Annotated

AiralogyMarkdown = Annotated[
    str,
    Field(
        description="An Airalogy Markdown string",
        json_schema_extra={"airalogy_type": "AiralogyMarkdown"},
    ),
]
"""
Airalogy's built-in type: AiralogyMarkdown.

Airalogy Markdown (AIMD) is a Markdown-like format used in Airalogy for writing and rendering text. It is designed to be user-friendly and easy to read, while also being powerful enough to support a wide range of formatting options. 

Strings of this type are parsed and rendered according to the Airalogy Markdown specification.

We explicitly refer to the format as Airalogy Markdown—rather than simply Markdown, because "Markdown" is a broad term: different platforms interpret it with their own extensions and syntax rules. Without a clear label, users may be unsure which rules to follow. Declaring Airalogy Markdown as the default Markdown flavor in Airalogy makes the applicable syntax and standards unambiguous.
"""



"""
This module contains the built-in types for Airalogy. These types could be used to define the Airalogy Protocol Model.

.. deprecated::
   `airalogy.built_in_types` is deprecated. Use `airalogy.types` instead.
"""

import warnings

# Re-export everything from the original modules for backward compatibility
from ..models.record import RecordId
from .recommended import Recommended
from .current_time import CurrentTime
from .file import (
    FileIdCSV,
    FileIdDOCX,
    FileIdJPG,
    FileIdJSON,
    FileIdMD,
    FileIdMP3,
    FileIdMP4,
    FileIdPDF,
    FileIdPNG,
    FileIdPPTX,
    FileIdAIMD,
    FileIdSVG,
    FileIdTIFF,
    FileIdTXT,
    FileIdWEBP,
    FileIdXLSX,
)
from .aimd import AiralogyMarkdown
from .user_name import UserName
from .ignore import IgnoreStr


# Issue deprecation warning
warnings.warn(
    "airalogy.built_in_types is deprecated and will be removed in a future version. "
    "Please use airalogy.types instead.",
    DeprecationWarning,
    stacklevel=2,
)

__all__ = [
    "UserName",
    "CurrentTime",
    "AiralogyMarkdown",
    "RecordId",
    "FileIdPNG",
    "FileIdJPG",
    "FileIdSVG",
    "FileIdWEBP",
    "FileIdTIFF",
    "FileIdMP4",
    "FileIdMP3",
    "FileIdAIMD",
    "FileIdMD",
    "FileIdTXT",
    "FileIdCSV",
    "FileIdJSON",
    "FileIdDOCX",
    "FileIdXLSX",
    "FileIdPPTX",
    "FileIdPDF",
    "Recommended",
    "IgnoreStr",
]



def all_var_ids_in_records(records: list[dict], var_ids: list[str]) -> bool:
    if not records:
        raise ValueError("The records list cannot be empty.")
    if not var_ids:
        raise ValueError("The var_ids list cannot be empty.")
    var_ids_set = set(var_ids)
    for record in records:
        try:
            var_dict = record["data"]["var"]
            if not isinstance(var_dict, dict):
                raise ValueError("The 'var' section must be a dictionary.")
        except (KeyError, TypeError) as exc:
            raise ValueError(
                "Each record must have a 'data' dict containing a 'var' dict."
            ) from exc
        if not var_ids_set.issubset(var_dict.keys()):
            return False
    return True



"""
Model for the Research Record.
"""

from datetime import datetime
from typing import Annotated

from pydantic import BaseModel, Field

RecordId = Annotated[
    str,
    Field(
        description="ID of an Airalogy Record. This ID is globally unique for each record in the Airalogy platform.",
        examples=["airalogy.id.record.xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx"],
        json_schema_extra={
            "airalogy_built_in_type": "RecordId",
        },
    ),
]


class RecordMetadata(BaseModel):
    """
    Model for the metadata of a Research Record.
    """

    initial_submission_user_id: str = Field(
        description="ID of the user who initially submitted the record.",
        examples=["zhang_san"],
    )
    initial_submission_time: datetime = Field(
        description="Time when the record was initially submitted.",
        examples=["2024-01-01T00:00:00+08:00"],
    )
    current_submission_user_id: str = Field(
        description="ID of the user who submitted the record at the current version.",
        examples=["zhang_san"],
    )
    current_submission_time: datetime = Field(
        description="Time when the record was submitted at the current version.",
        examples=["2024-01-01T00:00:00+08:00"],
    )
    lab_id: str = Field(
        description="Airalogy Lab ID. This ID is globally unique for each lab in the Airalogy platform.",
        examples=["airalogy"],
    )
    project_local_id: str = Field(
        description="Airalogy Project Local ID. This ID is unique within the lab. `project_local_id` is used to identify the project that the record belongs to.",
        examples=["airalogy_dev"],
    )
    rn_local_id: str = Field(
        description="Airalogy Research Node Local ID. This ID is unique within the project. `rn_local_id` is used to identify the research node that the record belongs to.",
        examples=["rn_test"],
    )
    rn_ver: str = Field(
        description="Research Node Version. `rn_ver` is used to identify the version of the Research Node that the record was recorded under.",
        examples=["0.1.0"],
    )
    record_num: int = Field(
        description="Record Number. `record_num` is used to identify the record's submission order under the Research Node. `1` means the first record under the Research Node.",
        examples=[1],
    )
    record_ver: int = Field(
        description="Record Version: `record_ver` is used to identify the edit submission number within the Research Node. `1` indicates the first submission. Starting from `2`, it indicates a submission that is edited based on the previous submission.",
        examples=[1],
    )
    sha1: str = Field(
        description="SHA-1 hash value of the `data` field of the record.",
    )


class RecordData(BaseModel):
    """
    Model for the `data` field of a Research Record.
    """

    rv: dict = Field(
        description="The data under the Research Variable template.",
    )
    rs: dict = Field(
        description="The data under the Research Step template.",
    )
    rc: dict = Field(
        description="The data under the Research Checkpoint template.",
    )


class Record(BaseModel):
    """
    Model for a Research Record. It primarily consists of 3 fields:
    - `record_id`: The ID of the record.
    - `metadata`: The metadata of the record.
    - `data`: The data of the record.
    """

    id: RecordId
    metadata: RecordMetadata
    data: RecordData



from typing_extensions import Annotated

from pydantic import Field

__all__ = ["IgnoreStr"]


IgnoreStr = Annotated[
    str,
    Field(
        default="",
        json_schema_extra={"airalogy_built_in_type": "IgnoreStr"},
        description="In an Airalogy Field of this type, a user may provide a string; however, when the Airalogy Record is submitted, the string is not persisted. This is intended for sensitive input (such as an API key) that the Airalogy Protocol Assigner can consume during processing but that should never be stored in the final Record.",
    ),
]



"""
Built-in type: CurrentTime
"""

from datetime import datetime

from pydantic import Field
from typing_extensions import Annotated

CurrentTime = Annotated[
    datetime,
    Field(
        description="The current time. The timezone is based on the user's browser.",
        json_schema_extra={"airalogy_type": "CurrentTime"},
    ),
]

CurrentProtocolId = Annotated[
    str,
    Field(
        description="The ID of the current Airalogy Protocol.",
        json_schema_extra={"airalogy_type": "CurrentProtocolId"},
        examples=[
            "airalogy.id.lab.my_lab.project.my_project.protocol.my_protocol.v.0.1.0"
        ],
    ),
]

CurrentRecordId = Annotated[
    str,
    Field(
        description="The ID of the current Airalogy Record.",
        json_schema_extra={"airalogy_type": "CurrentRecordId"},
        examples=["airalogy.id.record.12345678-1234-1234-1234-1234567890ab.v.1"],
    ),
]
